
#include <vector>
#include <assert.h>
#include "CACHE.hpp"
#include "Triplet.hpp"
#include "CS_ARRAY.hpp"

template< typename T >
class CS_TS {

   private:
	
	/** Convenience typedef */
	typedef unsigned long int ULI;

   protected:

	/** Number of non-zeros. */
	ULI nnz;

	/** Number of rows. */
	ULI nor;

	/** Number of columns. */
	ULI noc;

	/** Raw array storing the non-zeros and their locations. */
	//Triplet< T >* nzs;
	CS_ARRAY< T >* nzs;
	CS_ARRAY< ULI >* row;
	CS_ARRAY< ULI >* col;

	/** Which element is considered zero. */
	T zero_element;

	/* version control for cache resets */
	unsigned long int FAC;

   public:

	/** Base constructor. */
	CS_TS( std::vector< Triplet< T > > input, ULI m, ULI n, T zero ) {
		ULI offset = 0;

		zero_element = zero;
		nor = m;
		noc = n;
		nnz = input.size();
	//	nzs = new Triplet< T >[ nnz ];
		nzs = new CS_ARRAY< T >( nnz, "Non-zero array in CS_TS" );
		row = new CS_ARRAY< ULI >( nnz, "Row-index array in CS_TS" );
		col = new CS_ARRAY< ULI >( nnz, "Column-index array in CS_TS" );
		for( ULI i=0; i<nnz; i++ )
			if( input[ i ].value != zero_element ) {
				nzs->access( i - offset ) = input[ i ].value;
				row->access( i - offset ) = input[ i ].i();
				col->access( i - offset ) = input[ i ].j();
			} else
				offset++;
		nnz -= offset;
		FAC  = 0;
	}

	/** Basic MV. Performs z=Ax. */
	T* MV( T* x ) {
		T* ret = new T[ nor ]; CACHE::getInstance()->access( ret, sizeof( void* ), FAC );
		for( ULI i=0; i<nor; i++ ) {
			CACHE::getInstance()->access( &( ret[ i ] ), sizeof( T ), FAC );
			ret[ i ] = zero_element;
		}
		for( ULI i=0; i<nnz; i++ ) {
			/*CACHE::getInstance()->access( &( nzs[i].column ), sizeof( unsigned long int ) );
			CACHE::getInstance()->access( &( x[ nzs[i].j() ] ), sizeof( T ) );
			CACHE::getInstance()->access( &( nzs[i].value ), sizeof( T ) );
			CACHE::getInstance()->access( &( nzs[i].row ), sizeof( unsigned long int ) );
			CACHE::getInstance()->access( &( ret[nzs[i].i()] ), sizeof( T ) );*/
			
			assert( row->unrecorded_access( i ) >= 0 );
			assert( row->unrecorded_access( i ) <  nor );
			assert( col->unrecorded_access( i ) >= 0 );
			assert( col->unrecorded_access( i ) <  noc );
			ret[ row->const_access( i ) ] += nzs->const_access( i ) * x[ col->const_access( i ) ];
			CACHE::getInstance()->access( &( x[ col->unrecorded_const_access( i ) ] ), sizeof( unsigned long int ), FAC );
			CACHE::getInstance()->access( &( ret[ row->unrecorded_const_access( i ) ] ), sizeof( T ), FAC );
		}
		return ret;
	}

	/** Base destructor. */
	~CS_TS() {
		delete nzs;
		delete row;
		delete col;
	}

};


